header\<open>The Theory of Binary Decision Trees (BDT) \<close>

theory   BDT
imports  Interpretation
begin

section\<open>Basic structures and notations\<close>

type_synonym atom = bool

datatype     bdt  = Atom  atom  | IF  var bdt bdt

text \<open> Notations \<close>

syntax        "_topt"  :: "bdt"   ("topt")
translations  "topt" == "(CONST Atom) (CONST True)"

syntax        "_bott"        :: "bdt"   ("bott")
translations  "bott" == "(CONST Atom) (CONST False)"

definition    var   where "var n = IF n topt bott"


subsection \<open> Semantics \<close>

fun "interpret":: "interpretation \<Rightarrow> bdt \<Rightarrow> bool"
where "interpret I (Atom a)   =  a"
     |"interpret I (IF x P Q) =  (if I x then interpret I P else interpret I Q)"

(** tests *)

definition Pt where "Pt \<equiv> (IF (0::nat) (IF 1 topt topt) (IF 1 topt topt))"

value "interpret Itrue Pt"
value "(interpret Ifalse Pt)"
value "(interpret (interpl ([True,False])) Pt)"
value "(interpret (interpl ([False,True])) Pt)"

(* Trivial equality to be used for rewriting *)

lemma interpret_If : 
"interpret I (IF x P Q) = (if I x then interpret I P else interpret I Q)" by simp

lemma interpret_true [simp]: "I x \<Longrightarrow> interpret I (IF x P Q) = interpret I P" by simp

lemma interpret_false [simp]: "\<not>I x \<Longrightarrow> interpret I (IF x P Q) = interpret I Q" by simp

section\<open> Logical operations on BDT \<close>

fun    neg :: "bdt \<Rightarrow> bdt"
where "neg (Atom a) =  Atom (\<not> a)"
     |"neg (IF x l r) = IF x (neg l) (neg r)"



lemma negt_corr : "interpret I (neg P) = (\<not> (interpret I P))"
apply(rule bdt.induct)
 apply simp
apply simp
done

lemma negt_corr' : "interpret I (neg P) = (\<not> (interpret I P))"
proof(induct P) 
   case (Atom x) 
        show ?case by simp
next
   case (IF v P1 P2) 
        assume hyp1: "interpret I (neg P1) = (\<not> interpret I P1)" 
          and  hyp2: "interpret I (neg P2) = (\<not> interpret I P2)"
        show ?case by (simp add: hyp1 hyp2)
qed

(* structured, but even shorter *)

lemma negt_corr'' : "interpret I (neg P) = (\<not> (interpret I P))"
proof(induct P) 
   case (Atom x) show ?case by simp
next
   case (IF v P1 P2) then show ?case by simp
qed



section\<open>BinaryCombination.\<close>

fun comb where
    comb_a_a  :"comb opn (Atom a) (Atom b)  = Atom (opn a b)"
   |comb_a_if :"comb opn (Atom a) (IF x l r) = IF x (comb opn (Atom a) l) (comb opn (Atom a) r)"
   |comb_if_a :"comb opn (IF x l r) (Atom b) = IF x (comb opn l (Atom b)) (comb opn r (Atom b))"
   |comb_if_if:"comb opn (IF x l r) (IF y l' r') = 
                  (if x < y then  IF x (comb opn l (IF y l' r')) (comb opn r (IF y l' r'))
                            else if y < x then IF y (comb opn (IF x l r) l')(comb opn (IF x l r) r')
                                          else IF x (comb opn l l') (comb opn r r'))"

(** 
From Coq:
the following intuitive definition is not directly accepted 
Fixpoint bint (t u: bdt) : bdt := 
  match t with 
     Atomt a =>  match u with Atomt b => Atomt (op a b) 
                           |  IFt x l r => IFt x (bint t l) (bint t r)
                 end
   | IFt x l r => match u with Atomt b => IFt x (bint l u) (bint r u) 
                             |  IFt y l' r' => match nat_compare x y with 
                                 Lt => IFt x (bint l u) (bint r u) 
                                |Eq => IFt x (bint l l') (bint r r') 
                                |Gt => IFt y (bint t l') (bint t r') end
                  end
  end.
*)

(* This works in Isabelle without splitting in two fixpoints. 
   We use the nicer notation comb for bint, and we continue to ommit the ugly indexes t 
   which were not indexed after all.
*)


lemma comb_eq : "comb opn (IF x l r) (IF x l' r') = (IF x (comb opn l l') (comb opn r r'))"
by simp

lemma comb_le : "x < y \<Longrightarrow> comb opn (IF x l r) (IF y l' r') 
                           = (IF x (comb opn l (IF y l' r')) (comb opn r (IF y l' r')))"
by simp

lemma bint_ge : "y < x \<Longrightarrow> comb opn (IF x l r) (IF y l' r')
                           = IF y (comb opn (IF x l r) l') (comb opn (IF x l r) r')"
by simp


(** Instantiation on standard boolean operations *)
definition "and"  where "and  = comb (op \<and>)"
definition or     where "or   = comb (op \<or>)"
definition impl   where "impl = comb (op \<longrightarrow>)"


(** formula examples *)

definition p0   where "p0       = var 0"
definition p1   where "p1       = var 1"
definition P0   where "P0       = impl p0 p1"
definition P1   where "P1       = impl p1 P0"
definition dneg where "dneg P Q = impl (impl P Q) Q"
definition P2   where "P2       = dneg P0 p0"

value "P1"
value "P2"



lemma comb_correct : "interpret I (comb f P Q) = f (interpret I P) (interpret I Q)"
apply(rule_tac x=Q in spec)
apply(induct_tac P)
 apply(rule allI, rename_tac x Q, induct_tac Q)
 apply simp_all
apply(rename_tac "v" "x2" "x3")
apply(case_tac "I v", simp_all)
 apply(rule allI)
 apply(rename_tac "v" "x2" "x3" "Q") 
 apply(induct_tac Q)
  apply simp_all
 apply blast
apply(rule allI)
apply(rename_tac "v" "x2" "x3" "Q") 
apply(induct_tac Q)
 apply simp_all
by blast

text\<open> ... and the same as structured proof \<close>

lemma comb_correct' : "interpret I (comb f P Q) = f (interpret I P) (interpret I Q)"
proof(induct P arbitrary: Q)
   case (Atom a) show ?case by(induct_tac Q, simp_all)
next
   fix v::nat fix P1 P2 Q
   assume hyp1 : "\<And>Q. interpret I (comb f P1 Q) = f (interpret I P1) (interpret I Q)"
   and    hyp2 : "\<And>Q. interpret I (comb f P2 Q) = f (interpret I P2) (interpret I Q)"
   show "interpret I (comb f (IF v P1 P2) Q) = f (interpret I (IF v P1 P2)) (interpret I Q)"
        (is "?lhs = ?rhs")
        txt\<open>this is a pattern-matching against the above ``show'' defining @{term "?lhs"} \<close>
      apply simp 
      proof(cases "I v", simp_all) txt\<open> you might try here @{command print_cases} to check the
                                        environment of local assumptions here ... \<close>
         case True show "?lhs = f (interpret I P1) (interpret I Q)"
            txt \<open> the fact @{thm True} is here just a name for the assumption @{term "I v"} \<close>
            apply(induct Q) 
            apply(simp_all add: hyp1 hyp2 True)  
            using True by blast
      next 
         case False show "?lhs = f (interpret I P2) (interpret I Q)"
            apply(induct Q) 
            apply(simp_all add: hyp1 hyp2 False)  
            using False by blast            
      qed
qed 
text\<open> The advantages of this format are:
      \<^enum> proof and sub-proof structures were made explicit; 
        if one sub-proof breaks during development, the other 
        sub-proofs may still be valid. Structured proofs are therefore
        more stable under development as apply-style proofs, allow for a
        faster impact-analysis under change and
        can be evaluated in parallel --- the user profits even in short-term.
      \<^enum> The Isabelle community therefore encourages structured proofs.
      \<^enum> Major recurring proof patterns (induction with generalization,
        case-distinctions etc.) were made explicit which makes this
        proof format readable even for non-Isabelle users. (Well, sort of...)
      \<^enum> The notational burden to re-declare (abstractions of) the current proof state
        when entering a sub-proof can be alleviated by using case-notation and
        patterns from previous re-declarations. Note that the order of re-declaration
        is arbitrary. 
\<close>


lemma or_correct   : "interpret I (or P Q) = ((interpret I P) \<or> (interpret I Q))"
by (simp add: BDT.or_def comb_correct)

lemma and_correct  : "interpret I (and P Q) = ((interpret I P) \<and> (interpret I Q))"
by (simp add: BDT.and_def comb_correct)

lemma impl_correct : "interpret I (impl P Q) = ((interpret I P) \<longrightarrow> (interpret I Q))"
by (simp add: BDT.impl_def comb_correct)


primrec    exist\<^sub>v\<^sub>a\<^sub>r :: "var \<Rightarrow> bdt \<Rightarrow> bdt"
where     "exist\<^sub>v\<^sub>a\<^sub>r x (Atom a) = (Atom a)" (* x does not occur in t *)
         |"exist\<^sub>v\<^sub>a\<^sub>r x (IF y l r) = (if x < y then (IF y l r) (* x does not occur in t *) 
                                            else if y > x then IF y (exist\<^sub>v\<^sub>a\<^sub>r x l) (exist\<^sub>v\<^sub>a\<^sub>r x r)
                                                          else or l r)"

text\<open>Hint: ``fun'' does not work here (at least not without providing a non-trivial termination
     ordering) since it chooses the lexicographic ordering as default ... \<close>

(** ordered n i t : 
   all variables x in t are increasing on branches and i <= x < n *)

inductive          ordered :: "nat \<Rightarrow> nat \<Rightarrow> bdt \<Rightarrow> bool" 
where     Oatom : "ordered n i (Atom a)"
        | Oif   : "i <= x \<and> x < n 
                   \<Longrightarrow> ordered n (Suc x) l
                   \<Longrightarrow> ordered n (Suc x) r 
                   \<Longrightarrow> ordered n i (IF x l r)"


find_theorems name:"ordered" name:"BDT"  (* just to find out what the effect of "inductive" is ... *)
(* In Coq this was: 
Inductive ordered (n:nat) : nat -> bdt -> Prop := 
    Oatomt : forall i a, ordered n i (Atomt a)
  | Oift : forall i x l r, (i <= x < n) -> ordered n (S x) l -> ordered n (S x) r 
              -> ordered n i (IFt x l r).
*)


lemma neg_ordered_inv : "ordered n i P \<Longrightarrow> ordered n i (neg P)"
apply(erule  BDT.ordered.induct)
 apply simp apply(blast intro: Oatom)
apply simp apply(blast intro: Oif) 
done


lemma "ordered n i P \<Longrightarrow> ordered n i Q \<Longrightarrow> ordered n i (comb f P Q)"
apply(erule  BDT.ordered.induct)
 apply(erule  BDT.ordered.induct)
  apply simp apply(blast intro: Oatom)
sorry

(*
Lemma bint_ord : forall f i n P, ordered n i P -> forall Q, ordered n i Q
                 -> ordered n i (bint f P Q).
Proof.
induction 1.
- induction 1.
  + simpl.
    auto.
  + rewrite bint_eq.
    auto. 
- induction 1.
  + rewrite bint_eq; auto.
  + rewrite bint_eq.
    destruct (nat_compare_spec x x0).
    * subst x; auto.
    * auto with zarith.
    * auto with zarith.
Qed.

What precisely is "induction 1" ?
*)

lemma "ordered n i t \<Longrightarrow> (x < i \<or> n <= x) \<Longrightarrow>  interpret (update I x b) t = interpret I t"
thm BDT.ordered.induct 
apply(erule  BDT.ordered.induct) 
 apply simp
apply(simp add: update_neq update_eq)
sorry

(*
Lemma ordered_update : forall I x b i n t, ordered n i t -> (x < i \/ n <= x) ->
      interpt (update I x b) t = interpt I t.
induction 1.
-simpl.
 trivial.
-simpl.
intros.
rewrite update_neq; try omega.
rewrite IHordered1; try omega.
rewrite IHordered2; try omega.
trivial.
Qed.
*)

lemma exists_correct : 
"ordered n i t \<Longrightarrow> interpret I (exist\<^sub>v\<^sub>a\<^sub>r x t) 
                   = ( (interpret (update I x True) t) \<or> (interpret (update I x False) t))"
apply(erule  BDT.ordered.induct) 
 apply simp
apply simp
sorry

(*
Lemma existt_corr : forall i n I (x:var) (t:bdt), ordered n i t ->
  interpt I (existt x t) = orb (interpt (update I x true) t) (interpt (update I x false) t).
induction 1.
- simpl.
  auto.
- simpl.
destruct (nat_compare_spec x x0).
 + rewrite update_eq; trivial.
   rewrite update_eq; trivial.
   rewrite ort_corr.
   rewrite ordered_update with (1:=H0); try omega.
   rewrite ordered_update with (1:=H1); try omega.
   trivial.
 + rewrite update_neq; try omega.
   rewrite update_neq; try omega.
   do 2 (rewrite ordered_update with (1:=H0); try omega).
   do 2 (rewrite ordered_update with (1:=H1); try omega).
   simpl.
   destruct (I x0); auto.
 + rewrite update_neq; try omega.
   rewrite update_neq; try omega.
   simpl.
   destruct (I x0); auto.
Qed.
*)

(** * Using BDT to encode sets of states and do symbolic model-checking *)

(** We start from a (finite) problem where states depend on a set of boolean variables x0..x(n-1) 
    Each state is represented as a formula over these variables corresponding to a bdt ordered between 
    0 and n.
    Each set of states can also be represented by a formula over the same variables 
    (takes the disjunction of all sets)

    The transition system is represented by a formula with 2n variables corresponding to the values 
    before (x_i) and after (x_(n+i)). 
       - A single transition between state Gi(x_0,..,x_(n-1)) and Gj(x_0,..,x_(n-1))
    becomes a formula T_(ij)(x_0,..,x_(2n-1))=G1(x_0,..,x_(n-1))/\G2(x_n,..,x_(2n-1))
    The transition system is a formula T(x_0,..,x_(2n-1))
    corresponding to the disjonction of all transitions.

    We can now compute a set of states leading to a set of situations R in k or less steps
        R_0=R
        R_(i+1)(x_0,..x_(n-1)) = 
          R_i(x_0,..x_(n-1))\/exists xn,..,x_(2n-1), T(x_0,..,x_(2n-1))/\R_i(x_n,..x_(2n-1))
    We will reach a fixpoint after a finite number of steps that is a solution of our problems.

    Of course the efficiency of this method depends on a more efficient representation of bdt 
    using reduced bdd,  sharing mechanisms and computation with memoization.
*) 

end
